{
   muu Library for TMT Pascal - MTRR support unit
   Copyright (C) 2000 Tor Arne Gustad (anden@online.no)

   This source code is licensed under the GNU Public License.
   See the GNU General Public License for more details.
}

UNIT MTRR;

INTERFACE
{$R-,I+,Q-}

    // MTRR memory types
CONST mtrr_Uncacheable =     $000;   // Uncacheable
      mtrr_WriteCombine =    $001;   // Write-Combine
      mtrr_WriteThrough =    $004;   // Write-Through
      mtrr_WriteProtected =  $005;   // Write-Protected
      mtrr_WriteBack =       $006;   // Write-Back


    // User interface
    // -------------------
    // phys_base = physical base of memory location
    // phys_size = physical size of memory
    // mem_type = MTRR memory type (= mtrr_* as seen above)

PROCEDURE mtrr_enable(phys_base,phys_size:dword;mem_type:byte);
PROCEDURE mtrr_disable(phys_base,phys_size:dword;mem_type:byte);


IMPLEMENTATION

        // instruction opcodes (byte-swapped)
CONST _rdmsr = $320F;
      _wrmsr = $300F;
    
        // internal constants
      _n_mtrr = 7;          // # of MTRRs
      _mtrr_u4mask = $0F;   // upper 4 bits of physMask (36bit variable)

        // MTRR mask settings
      _mtrrValid = $800;        // enable register pair

        // MSRs
      _mtrrdefType =    $2FF;
      _mtrrphysBase =   $200;   // base0, add (mtrr*2) for mtrr[0..7]
      _mtrrphysMask =   $201;   // mask0, add (mtrr*2) for mtrr[0..7]

    // MSR qword type
TYPE _msr = ARRAY[0..1] OF dword;


    // Internal functions
    // --------------------

FUNCTION _read_msr(n_msr:dword): _msr;
  VAR _eax,_edx:    dword;
BEGIN
  ASM
    mov  ecx,  [n_msr]
    dw   _rdmsr

    mov  [_eax], eax
    mov  [_edx], edx
  END;

  result[0] := _eax;
  result[1] := _edx;
END;

PROCEDURE _write_msr(n_msr:dword;msr:_msr);
  VAR _eax,_edx:    dword;
BEGIN
  _eax := msr[0];
  _edx := msr[1];
  
  ASM
    mov  ecx,  [n_msr]
    mov  eax,  [_eax]
    mov  edx,  [_edx]

    dw   _wrmsr
  END;
END;



    // User interface functions
    // -------------------------

PROCEDURE mtrr_enable;
  VAR mtrrphysbase: _msr;
      mtrrphysmask: _msr;
      mtrrdefsave:  _msr;
      msrblank:     _msr;
      mtrr:         dword := 0;
BEGIN
    // calculate physbase/physmask
  mtrrphysbase[0] := phys_base OR mem_type;
  mtrrphysbase[1] := 0;

  mtrrphysmask[0] := ($FFFFFFFF-phys_size)+mem_type+_mtrrValid;
  mtrrphysmask[1] := _mtrr_u4mask;

    // try to locate a free mtrr
  DECLARE VAR base,mask:    _msr;
              count:        dword;
              mtrr_found:   boolean := false;
  BEGIN
    FOR count := 0 TO _n_mtrr DO
    BEGIN
      base := _read_msr(_mtrrphysBase+count*2);
      mask := _read_msr(_mtrrphysMask+count*2);

      IF (NOT mtrr_found) AND
         (base[0] = mtrrphysbase[0]) AND
         (base[1] = mtrrphysbase[1]) THEN
        IF (mask[0] = mtrrphysmask[0]) AND
           (mask[1] = mtrrphysmask[1]) THEN
            Exit
          ELSE BEGIN    // remap phys_base
            mtrr := count;
            mtrr_found := true;
          END;

      IF (NOT mtrr_found) AND
         (base[0] = 0) AND
         (base[1] = 0) THEN
        BEGIN   // free mtrr
          mtrr := count;
          mtrr_found := true;
        END;

      IF (mtrr_found) THEN
        break;
    END;

        // exit if none available
    IF (NOT mtrr_found) THEN
      Exit;
  END;

    // disable mtrr
  fillchar(msrblank,sizeof(_msr),0);
  mtrrdefSave := _read_msr(_mtrrdefType);
  _write_msr(_mtrrdefType,msrBlank);
  
    // map new mtrr
  _write_msr(_mtrrphysBase+mtrr*2,mtrrphysBase);
  _write_msr(_mtrrphysMask+mtrr*2,mtrrphysMask);

    // enable mtrr
  _write_msr(_mtrrdefType,mtrrdefSave);
END;


PROCEDURE mtrr_disable;
  VAR mtrrphysbase: _msr;
      mtrrphysmask: _msr;
      mtrrdefsave:  _msr;
      msrblank:     _msr;
      mtrr:         dword := 0;
BEGIN
    // calculate physbase/physmask
  mtrrphysbase[0] := phys_base OR mem_type;
  mtrrphysbase[1] := 0;

  mtrrphysmask[0] := ($FFFFFFFF-phys_size)+mem_type+_mtrrValid;
  mtrrphysmask[1] := _mtrr_u4mask;

    // try to locate a free mtrr
  DECLARE VAR base,mask:    _msr;
              count:        dword;
              mtrr_found:   boolean := false;
  BEGIN

    FOR count := 0 TO _n_mtrr DO
    BEGIN
      base := _read_msr(_mtrrphysBase+count*2);
      mask := _read_msr(_mtrrphysMask+count*2);

      IF (NOT mtrr_found) AND
         (base[0] = mtrrphysbase[0]) AND
         (base[1] = mtrrphysbase[1]) AND
         (mask[0] = mtrrphysmask[0]) AND
         (mask[1] = mtrrphysmask[1]) THEN
        BEGIN
          mtrr := count;
          mtrr_found := true;
        END;

      IF (mtrr_found) THEN
        break;
    END;

        // exit if none available
    IF (NOT mtrr_found) THEN
      Exit;
  END;

    // disable mtrr
  fillchar(msrblank,sizeof(_msr),0);
  mtrrdefSave := _read_msr(_mtrrdefType);
  _write_msr(_mtrrdefType,msrBlank);
  
    // clear mtrr
  _write_msr(_mtrrphysBase+mtrr*2,msrblank);
  _write_msr(_mtrrphysMask+mtrr*2,msrblank);

    // enable mtrr
  _write_msr(_mtrrdefType,mtrrdefSave);
END;

end.
